/**
 * MPQAudioWorkerPlayThread.java: Playback worker thread class.
 *            Copyright (C) 2012,2013 donfede@casagrau.org
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * created 2012-12-04 donfede@casagrau.org
 * credit to: Reto Meier, EMEA Android Developer Advocate, Google UK; for a record and playback example -- http://eurodev.blogspot.com/2009/09/raw-audio-manipulation-in-android.html
 *			  Stjepan Rajko, urbanSTEW -- (GPL) RehearsalAssistant/RehearsalAudioRecorder.java
 *			  cristi (Cristian Baita) -- http://indyvision.net/2010/02/android-threads-tutorial-part-3/
 *			  Ringdroid, with a primary developer of Dominic Mazzoni, a Google employee and also author of Audacity. --  http://code.google.com/p/ringdroid/
 * $Id: MPQAudioWorkerPlayThread.java,v 1.7 2013-03-24 16:07:44 donfede Exp $
 */

package org.casagrau.mpq.audio_service;

import java.io.File;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.text.SimpleDateFormat;
import java.util.Date;

import org.casagrau.mpq.mpq_android.SessionsDBProvider.SessionsDB;

import android.content.Context;
import android.media.AudioFormat;
import android.media.AudioManager;
import android.media.AudioTrack;
import android.os.Environment;
import android.os.Handler;
import android.util.Log;



public class MPQAudioWorkerPlayThread extends MPQAudioWorkerThread {
	private static final String LOGTAG = "MPQAudioWorkerPlayThread";

	/** 
	 * Constructor for the worker thread.
	 * {@link #MPQAudioWorkerThread}  
	 */
	public MPQAudioWorkerPlayThread(Handler parentHandler, String filename, long filepos, int state, Context context) {
		super(parentHandler, filename, filepos, state, context);
	}
	
	/**
	 * Play audio from a wave file
	 * Credit to: Reto Meier, EMEA Android Developer Advocate, Google UK -- http://eurodev.blogspot.com/2009/09/raw-audio-manipulation-in-android.html
	 * 	          Ringdroid, with a primary developer of Dominic Mazzoni, a Google employee and also author of Audacity. --  http://code.google.com/p/ringdroid/
	 * @param intent with the action and extra content to process. 
	 * 					Extra ARG_FILENAME is checked for to open.
	 * @throws IllegalStateException if there are errors writing the file
	 */
	public void run() { 
		Log.d(LOGTAG, "runPLAY() 00");
		// Get the file we want to playback.
		File lFile = new File(Environment.getExternalStorageDirectory().getAbsolutePath() + mFilename);
		Log.d(LOGTAG, "runPLAY() 06  file:" + lFile + "mStateThread: " + mStateThread);

		try {
			// Open file to play back
			mFileHandle = new RandomAccessFile(lFile, "r");

			int lfrequency = 0;			// Local sample bitrate.
			int lchannelConfiguration = 0;// Channel config code, ex: AudioFormat.CHANNEL_OUT_MONO .
	    	int lchannelCount = 0;		// Channel config count.
			int laudioEncoding = 0;		// Audio sample bit rate encoding, ex: AudioFormat.ENCODING_PCM_16BIT . 
	    	int lsampleBit = 0; 		// Audio sample bit rate, corresponding with laudioEncoding.
	    	boolean ldataFlag = false;	// Flag variable to track when the data chunk of the WAVE file is found.
	    	int lchunkLen = 0;			// Chunk size, obtained from headers.
			
			// Read WAVE header. -- credit: Dominic Mazzoni, RingDroid CheapWAV
	        long lFileSize = mFileHandle.length();
	        if (lFileSize < 128) {
	            throw new java.io.IOException("File too small to parse"); // FIXME: these exceptions get caught immediately right here...
	        }
	        int lOffset = 0;
	        byte[] header = new byte[12];
	        
	        mFileHandle.read(header, 0, 12);
	        lOffset += 12;
	        if (header[0] != 'R' ||
	            header[1] != 'I' ||
	            header[2] != 'F' ||
	            header[3] != 'F' ||
	            header[8] != 'W' ||
	            header[9] != 'A' ||
	            header[10] != 'V' ||
	            header[11] != 'E') {
	            throw new java.io.IOException("Not a WAV file");
	        }

			// Read headers loop
	        while ( (lOffset + 8 <= lFileSize) && (!ldataFlag) ) {
	            byte[] chunkHeader = new byte[8];
	            mFileHandle.read(chunkHeader, 0, 8);
	            lOffset += 8;

	            lchunkLen =
	                ((0xff & chunkHeader[7]) << 24) |
	                ((0xff & chunkHeader[6]) << 16) |
	                ((0xff & chunkHeader[5]) << 8) |
	                ((0xff & chunkHeader[4]));

	            if (chunkHeader[0] == 'f' &&
	                chunkHeader[1] == 'm' &&
	                chunkHeader[2] == 't' &&
	                chunkHeader[3] == ' ') {
	                if (lchunkLen < 16 || lchunkLen > 1024) {
	                    throw new java.io.IOException(
	                        "WAV file has bad fmt chunk");
	                }

	                byte[] fmt = new byte[lchunkLen];
	                mFileHandle.read(fmt, 0, lchunkLen);
	                lOffset += lchunkLen;

	                int lformat =
	                    ((0xff & fmt[1]) << 8) |
	                    ((0xff & fmt[0]));
	                lchannelCount =
	                    ((0xff & fmt[3]) << 8) |
	                    ((0xff & fmt[2]));
	                lfrequency =
	                    ((0xff & fmt[7]) << 24) |
	                    ((0xff & fmt[6]) << 16) |
	                    ((0xff & fmt[5]) << 8) |
	                    ((0xff & fmt[4]));
	                // 4n AvgBytesPerSec [11 -8]
	                // 2n BlockAlign [13-12]
	                lsampleBit = 
	                	((0xff & fmt[15]) << 8) |
	                	((0xff & fmt[14]));

	                if (lformat != 1) {
	                    throw new java.io.IOException(
	                        "Unsupported WAV file encoding");
	                }
	                
	    	    	if (1 == lchannelCount) lchannelConfiguration = AudioFormat.CHANNEL_OUT_MONO;
	    	    	else if (2 == lchannelCount) lchannelConfiguration = AudioFormat.CHANNEL_OUT_STEREO;
	    	    	else throw new IllegalStateException("unsuported channel count: " + lchannelCount);

	    	    	if (16 == lsampleBit) laudioEncoding = AudioFormat.ENCODING_PCM_16BIT;
	    	    	else if (8 == lsampleBit) laudioEncoding = AudioFormat.ENCODING_PCM_8BIT;
	    	    	else throw new IllegalStateException("unsuported sample bitrate: " + lsampleBit);
	            } else if (chunkHeader[0] == 'd' &&
	                       chunkHeader[1] == 'a' &&
	                       chunkHeader[2] == 't' &&
	                       chunkHeader[3] == 'a') {
	                if (lchannelCount == 0 || lfrequency == 0 || lsampleBit == 0) {
	                    throw new java.io.IOException("Bad WAV file: data chunk before fmt chunk");
	                }
	                ldataFlag = true;
	            } else {
	            	mFileHandle.seek(lchunkLen);
	                lOffset += lchunkLen;
	            }
	        } // END Read headers loop

     
	        long laudioStartByte = mFileHandle.getFilePointer();
	    	int lframeSize = (lsampleBit/8)*lchannelCount;
	    	Log.d(LOGTAG, "runPLAY() 20  framesize:" + lframeSize + " channelCount: " + lchannelCount + " sampleBit: " + lsampleBit + " dataStartByte: " + laudioStartByte);
			
	    	long lframeCount = lchunkLen / lframeSize;
			double lfrequencyMS = lfrequency / 1000.0; // cycles per millisecond
	    	long lfileMS = (long) (lframeCount / lfrequencyMS);
	    	Log.d(LOGTAG, "runPLAY() 24  fileMS: " + lfileMS + " frameCount: " + lframeCount + " lfreq: " + lfrequency + " lfreqMS: " + lfrequencyMS);
			sendPMsgLen(lfileMS);

			// Create a buffer for audio playback that is appropriate for input file.
			int lbufferSize = AudioTrack.getMinBufferSize(lfrequency, lchannelConfiguration, laudioEncoding);
			lbufferSize = 2 * lbufferSize;
			if (lbufferSize < 16384) lbufferSize = 16384; // FIXME: hack to increase buffer size (saw it at 2400) -- better solution would be to 'do math' and try and buffer $X milliseconds
			byte[] buffer = new byte[lbufferSize];
			Log.d(LOGTAG, "runPLAY() 30  bufSize(2xmin, or 16384): " + lbufferSize);
			// Create a new AudioTrack object using the same parameters as the AudioRecord
			// object used to create the file.
			AudioTrack laudioTrack = new AudioTrack(AudioManager.STREAM_MUSIC,
					lfrequency,
					lchannelConfiguration,
					laudioEncoding,
					lbufferSize,
					AudioTrack.MODE_STREAM);

			if (laudioTrack.getState() != AudioTrack.STATE_INITIALIZED)
				throw new IOException("runPLAY() AudioTrack initialization failed");

			laudioTrack.play();
			long lnowTime = System.currentTimeMillis();
			SimpleDateFormat formatter = new SimpleDateFormat("yyyy/MM/dd HH:mm:ss");
			String lstartTimeStr = formatter.format(new Date(lnowTime));
			Log.d(LOGTAG, "runPLAY() 50  start time" + lnowTime + "str start: " + lstartTimeStr);
			Log.d(LOGTAG, "runPLAY() 52 playbackNotificationRate:" + laudioTrack.getPositionNotificationPeriod() );
			
			int lbufferReadSize;	// Variable to store the size read into the buffer.
			@SuppressWarnings("unused")
			int lplaySize;			// Variable to store the size played to the speakers/AudioTrack.
			long lFileposBytes = 0;
		
			// Convert mFilepos into bytes and seek. // FIXME: this initial call uses previous values stored by the service; consider using the temporary mFileposSeek here...
			if (mFilepos > 0) {
				lFileposBytes = cursorMSToBytes(mFilepos, lframeSize, lfrequency);	
				Log.w(LOGTAG, "runPLAY() 61 mFilepos: " + mFilepos + " lFileposBytes: " + lFileposBytes + " FH.FilePointer: " + mFileHandle.getFilePointer() + " audiostartbyte: " + laudioStartByte);
	        	mFileHandle.seek(laudioStartByte + lFileposBytes);
				Log.w(LOGTAG, "runPLAY() 63 FH.FilePointer: " + mFileHandle.getFilePointer());
			}			
			sendPMsgPos(mFilepos, false);
		
			// Playback loop
			while ( (lFileposBytes < lchunkLen) && ((MPQAudioService.FLG_SRV_STATE_PLAY == mStateThread) || 
												(MPQAudioService.FLG_SRV_STATE_PAUSE_PLAY == mStateThread)) ) {
				Log.d(LOGTAG, "runPLAY() 70  IN loop read -  headPos:" + laudioTrack.getPlaybackHeadPosition() + " mStateThread: " + mStateThread);
				//xxLog.d(LOGTAG, "runPLAY() 72             now time:" + formatter.format(new Date(lnowTime)) );
				Log.w(LOGTAG, "runPLAY() 72AA           fileposbytes: " + lFileposBytes + " chunkLen: " + lchunkLen + " mFilepos: " + mFilepos);
				
				if (MPQAudioService.FLG_SRV_STATE_PLAY == mStateThread) {
					if (AudioTrack.PLAYSTATE_PLAYING != laudioTrack.getPlayState()  ) 
						laudioTrack.play();			
					
					lbufferReadSize = mFileHandle.read(buffer);
					//xxLog.d(LOGTAG, "runPLAY() 73A        IN loop  PLAY read: " + lbufferReadSize + " buf says:" + buffer.length + " headPos:" + laudioTrack.getPlaybackHeadPosition() + " mStateThread: " + mStateThread);
					//xxLog.d(LOGTAG, "runPLAY() 73B        IN loop  PLAY filepos: " + mFileHandle.getFilePointer() + " file len: " + mFileHandle.length() );

					if (lbufferReadSize > 0) {
						// Write the music buffer to the AudioTrack object
						// FIXME: i fear i should have a ring buffer here (and read, separately from the audio playback...)
						lplaySize = laudioTrack.write(buffer, 0, lbufferReadSize);
						lFileposBytes = lFileposBytes + lbufferReadSize;
						mFilepos = cursorByteToMS(lFileposBytes, lframeSize, lfrequency);
						sendPMsgPos(mFilepos, false);
						//xxLog.d(LOGTAG, "runPLAY() 74 wrote playSize: " + lplaySize + " headPos:" + laudioTrack.getPlaybackHeadPosition() + " filepos: " + mFilepos );
						//xxLog.d(LOGTAG, "runPLAY() 75 wrote filepos: " + mFilepos + " filepos STR: " + SessionsDB.longSessionMSecString(mFilepos));
					} else 
						break; // We've likely hit EOF.  // FIXME: should probably actually check for EOF... (but we're stopping either way...)
				} else { // FLG_SRV_STATE_PAUSE_PLAY
					if (AudioTrack.PLAYSTATE_PAUSED != laudioTrack.getPlayState()  ) 
						laudioTrack.pause();
					// wait before checking again.
					try {
						Thread.sleep(AUDIO_PAUSE_MS);
					} catch (Exception e) {
						e.getLocalizedMessage();
					}	
					//xxLog.d(LOGTAG, "recordWAV() 77      IN while PAUSE WAIT mStateThread: " + mStateThread + " nowTime:" + formatter.format(new Date(lnowTime)) );            		
				}

				// If a seek request is found, convert mFilepos into bytes and seek.
				if (mFileposSeek >= 0) {
					mFilepos = mFileposSeek;
		        	mFileposSeek = -1; // Reset seek flag.
					lFileposBytes = cursorMSToBytes(mFilepos, lframeSize, lfrequency);
					Log.w(LOGTAG, "runPLAY() 82 mFilepos: " + mFilepos + " lFileposBytes: " + lFileposBytes + " FH.FilePointer: " + mFileHandle.getFilePointer() + " audiostartbyte: " + laudioStartByte);
		        	mFileHandle.seek(laudioStartByte + lFileposBytes);
					Log.w(LOGTAG, "runPLAY() 84 FH.FilePointer: " + mFileHandle.getFilePointer());

				}
				Log.w(LOGTAG, "runPLAY() 88 lFileposBytes: " + lFileposBytes + " lchunkLen: " + lchunkLen + " mStateThread: " + mStateThread);
				
				lnowTime = System.currentTimeMillis();
			} // END while() playback loop
			Log.d(LOGTAG, "runPLAY() 90 OUT playing while now/end time:" + formatter.format(new Date(lnowTime)) );
			laudioTrack.stop();
			laudioTrack.release();

			// Set state to stopped, and send updates.
			mFilepos = 0;
			sendPMsgPos(mFilepos, true);
			if (MPQAudioService.FLG_SRV_STATE_STOP != mStateThread) {		
	    		Log.d(LOGTAG, "runPLAY() 94 send startService intent STOP");
	    		sendSvcIntentStop();
			}
    		Log.d(LOGTAG, "runPLAY() 97 send startService intent OPEN");
    		sendSvcIntentOpen();
		} catch (Exception e) {
			Log.e(LOGTAG,"runPLAY() Playback Failed e:" + e.getLocalizedMessage());
		}
		Log.d(LOGTAG, "runPLAY() 99");
	} // END runPLAY()

} // END class MPQAudioWorkerPlayThread





