/**
 * SessionDetailFragment.java: Practice session detail, and recording.
 *            Copyright (C) 2012,2013 donfede@casagrau.org
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * created 2012-09-23 donfede@casagrau.org
 *
 * $Id: SessionDetailFragment.java,v 1.61 2013-03-27 22:23:03 donfede Exp $
 */

package org.casagrau.mpq.mpq_android;

import java.io.File;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;
import java.util.List;

import org.casagrau.mpq.audio_service.MPQAudioFragment;
import org.casagrau.mpq.audio_service.MPQAudioService;
import org.casagrau.mpq.mpq_android.SessionsDBProvider.SessionsDB;

import android.app.Activity;
import android.content.ComponentName;
import android.content.ContentUris;
import android.content.ContentValues;
import android.content.Context;
import android.content.Intent;
import android.content.ServiceConnection;
import android.content.SharedPreferences;
import android.database.Cursor;
import android.net.Uri;
import android.os.Bundle;
import android.os.IBinder;
import android.preference.PreferenceManager;
import android.support.v4.app.DialogFragment;
import android.support.v4.app.Fragment;
import android.support.v4.app.FragmentManager;
import android.support.v4.app.FragmentTransaction;
import android.support.v4.app.LoaderManager;
import android.support.v4.content.CursorLoader;
import android.support.v4.content.Loader;
import android.support.v4.widget.SimpleCursorAdapter;
import android.util.Log;
import android.view.LayoutInflater;
import android.view.View;
import android.view.View.OnClickListener;
import android.view.ViewGroup;
import android.widget.Button;
import android.widget.EditText;
import android.widget.Spinner;
import android.widget.TextView;
import android.widget.Toast;

/**
 * Fragment with GUI for Session Detail. This screen will be active while a user
 * is in a practice session, and allow the user to review and edit details about
 * a practice session.
 */
public class SessionDetailFragment extends Fragment implements
		LoaderManager.LoaderCallbacks<Cursor> , 
		MPQGenDialogFragment.MPQGenDialogListener {
	private static final String LOGTAG = "SessionDetailFragment";

	/*******************
	 * 	Public constants
	 */
	public static final String ARG_DETAIL_FLAG = "detail_session_flag";	// Argument bundle handle.
	public static final long ARG_DETAIL_FLAG_START = -1;				// Negative argument value code to add and start a new session.
	public static final long ARG_DETAIL_FLAG_CURRENT = 0;				// Zero value argument value code to show the current or most recent session.
																		// Positive value argument value code represent session index values to retrieve and display.
	
	/*******************
	 * Private constants
	 */
	private static final String BNDL_SESSION_ID = "session_id"; 			// Bundle handle to SesId
	private static final String BNDL_SESSION_STATE = "session_state"; 		// Bundle handle to SesState
	private static final String BNDL_SESSION_FILENAME = "session_filename";	// Bundle handle to FileName.  (donfede - this currently helps identify if we're recording or not; should probably drop this, and instead get info from bound audio_service . 
	
	private static final int DLG_SAVE_DISCARD = 10; 		// Identifier code for the save discard dialog.
	
	private static final int FLG_SES_STATE_VIEW = 0;		// Session state is in View mode.
	private static final int FLG_SES_STATE_EDIT = 1;		// Session state is in Edit mode.
	
	private static final int FLG_SES_ACTION_REC = -1;		// Session action flag, to record a new session.
	private static final int FLG_SES_ACTION_OPEN = 0;		// Session action flag, to open an existing session.
	
	/**
	 * Unique integer identifiers for Loaders.
	 */
	private static final int LDR_GOALS = 0;
	private static final int LDR_INSTRUMENTS = 1;
	private static final int LDR_SESSIONS = 2;

	/***************************
	 * Private member attributes
	 */
	// Adapters to bind to CursorLoaders
	private SimpleCursorAdapter mAdapterGoals;
	private SimpleCursorAdapter mAdapterInstruments;
	
	// FIXME: potential solution to static problem; saving in onPause() -- http://developer.android.com/guide/topics/data/data-storage.html#pref -- http://stackoverflow.com/questions/151777/saving-activity-state-in-android
//	private long mSesId = ARG_DETAIL_FLAG_CURRENT; // Identifier of the current practice session for which detail is being displayed and processed. // FIXME: keeping this static, to maintain user consistency when returning to the tab
	private static long mSesId = ARG_DETAIL_FLAG_CURRENT; // Identifier of the current practice session for which detail is being displayed and processed. // FIXME: keeping this static, to maintain user consistency when returning to the tab
//	private int mSesActionFlag = FLG_SES_ACTION_OPEN;	  // Flag variable to track if a session should be opened or created.
//	private long mSesActionId = 0;						  // Flag variable with session id associated with mSesIdActionFlag.
	private static int mSesActionFlag = FLG_SES_ACTION_OPEN;	  // Flag variable to track if a session should be opened or created.
	private static long mSesActionId = 0;						  // Flag variable with session id associated with mSesIdActionFlag.
	private int mSesState;	 // Flag to track session state (View, Edit, "Create"[not sure if latter is needed)
	private long mSesCount;	 // Number of rows of practice sessions.
	private long mSesMin;	 // Minimum row id value for practice sessions.
	private long mSesMax;    // Max row id value for practice sessions.
	private MPQMainTabActivity mAct;	// Local caching variable with the activity.
	/**
	 * Class to keep track of meta data associated with a UI widget element. 
	 */
	private class SessionDetailWidgetMeta {
		// Widget Type constants
		private static final int WT_INVALID = -1;
		private static final int WT_SPINNER = 0;
		private static final int WT_EDITTEXT = 1;
		private static final int WT_TEXTVIEW = 2;

		private String mDBCol; // for now - access the members directly from the calling object... try to restrict access to within this class using private (might be off here)
		private int mDBColPos;
		private int mResourceID;
		private View mUIWidget;
		private int mWidgetType; // maybe use Cursor.FIELD_TYPE_xxx constants ... but still need widget type... 
		//FIXME: should this object include : A)cursor method to extract? 
		/** 
		 * Constructor that populates empty values
		 */
		SessionDetailWidgetMeta() {
			mDBCol = "";
			mDBColPos = -1;
			mResourceID = -1;
			mUIWidget = null;
			mWidgetType = WT_INVALID;
		}
	} // END class SessionDetailWidgets
	// An array to keep track of widget-DB data. 
	private List<SessionDetailWidgetMeta> mDetailWidgets = new ArrayList<SessionDetailWidgetMeta>();
	private String mFileName = null;

	// Handle to bound audio service. 
	private MPQAudioService mAudioService;		// Handle to the audio service, received from a binding connection.
	private boolean mAudioServiceBound = false;	// Flag to track if the audio service is bound.


	// Handles to various UI widgets.
	private Button mPrevButton = null;
	private Button mNextButton = null;
	private Button mEditLockButton = null;

	
	/*
	 * Parameterless constructor, required by framework
	 */
	public SessionDetailFragment() {
		super();
	}

	@Override
	public void onAttach(Activity a) {
		super.onAttach(a);
		Log.d(LOGTAG, "onAttach() 00");
		Log.d(LOGTAG, "onAttach() 10 activity:" + a.getLocalClassName());
		mAct = (MPQMainTabActivity) a;
		
		// Bind to the audio service.
    	Log.d(LOGTAG, "onAttach() 30 - binding to audio service");
    	Intent intent = new Intent(mAct, MPQAudioService.class);
		intent.setAction(MPQAudioService.ACTION_BIND);
		mAct.bindService(intent, mAServiceConnection, Context.BIND_AUTO_CREATE);
    	Log.d(LOGTAG, "onAttach() 99");
	}
	
	@Override
	public void onCreate(Bundle savedInstanceState) {
		super.onCreate(savedInstanceState);
		Log.d(LOGTAG, "onCreate() 00");		
	}
	
	
	/**
	 * Create the view for this fragment.
	 * Fragment arguments are checked and handled.
	 * Audio sub-Fragment is created loaded.
	 * Fragment XML layout is loaded and returned.
	 * {@link #mSesId} is set with a new or existing practice session to be opened by methods called later.
	 * {@link #mSesActionFlag} and {@link #mSesActionId} are also set with code to OPEN or REC a session. 
	 */
	@Override
	public View onCreateView(LayoutInflater inflater, ViewGroup container,
			Bundle savedInstanceState) {
		super.onCreateView(inflater, container, savedInstanceState);
		Log.d(LOGTAG, "onCreateView() 00 mSesId:" + mSesId);		

		// Get the previous state, or initialize state to viewing.
		if (savedInstanceState != null) {
			Log.d(LOGTAG, "onCreateView() 10   found bundle:" + savedInstanceState.toString() );
			mSesState = savedInstanceState.getInt(BNDL_SESSION_STATE);
			mFileName = savedInstanceState.getString(BNDL_SESSION_FILENAME);
			Log.d(LOGTAG, "onCreateView() 11   found BNDL State:" + mSesState + " filename: " + mFileName);		
		} else {
			mSesState = FLG_SES_STATE_VIEW;
			Log.d(LOGTAG, "onCreateView() 14   init NO BNDL State:" + mSesState);
		}
		
		// Check Fragment arguments for flag value with session to display or create.
		long lflag = SessionDetailFragment.ARG_DETAIL_FLAG_CURRENT;
		if (getArguments() != null) {
			lflag = getArguments().getLong(ARG_DETAIL_FLAG, 
					SessionDetailFragment.ARG_DETAIL_FLAG_CURRENT); 
			Log.d(LOGTAG, "onCreateView() 20   found args lflag:" + lflag);		
		} else Log.d(LOGTAG, "onCreateView() 21   NO args   default lflag:" + lflag);
		
		if (lflag < 0 ) {
			// Open a new practice session.
			mSesId = newSessionDetailData();
			mSesActionFlag = FLG_SES_ACTION_REC;
			mSesActionId = mSesId;
			Log.d(LOGTAG, "onCreateView() 30   add new mSesId:" + mSesId);
		} else if (lflag > 0) {
			// (try to) Open specific session.
			mSesId = lflag;
			mSesActionFlag = FLG_SES_ACTION_OPEN;
			mSesActionId = mSesId;
			Log.d(LOGTAG, "onCreateView() 60   get try mSesId:" + mSesId + " state:" + mSesState);
		} else if (lflag == 0) {
			// Open current or most recent practice session.
			if (savedInstanceState != null) {
	        	mSesId = savedInstanceState.getLong(BNDL_SESSION_ID);				
			} // FIXME: if mSesId is 0, should we set it to max session value?
			mSesActionFlag = FLG_SES_ACTION_OPEN;
			mSesActionId = mSesId;
			Log.d(LOGTAG, "onCreateView() 70   savedInstance not null mSesId:" + mSesId + " state:" + mSesState);		
        } 
		
		// Sanity check that mSesId refers to a valid session; alternately get Max.
		int i = 0;
		if (mSesId > 0)
			i = getSessionDetailIdCount(mSesId);
		if ( (i != 1) || (mSesId == 0) ) {
			updateSessionsCount();
			mSesId = mSesMax;
			mSesActionFlag = FLG_SES_ACTION_OPEN;
			mSesActionId = mSesId;
	    }
		Log.d(LOGTAG, "onCreateView() 76   mSesId: " + mSesId + " mSesActionId: " + mSesActionId + " mSesAF: " + mSesActionFlag);		

		
		// Load sub-Fragment with audio UI.
		Log.d(LOGTAG, "onCreateView() 80 audio fragment load start");
    	String fragStr = getString(R.string.frag_audio_ui);     
    	int fragView = R.id.frag_audio;
    	FragmentManager fm = getChildFragmentManager();
		MPQAudioFragment mpqaf = (MPQAudioFragment)  fm.findFragmentByTag(fragStr);
		if (null == mpqaf) {
			Log.d(LOGTAG, "onCreateView() 87 null audio fragment - creating new");
			FragmentTransaction fmt = fm.beginTransaction();
			mpqaf = new MPQAudioFragment();
			fmt.replace(fragView, mpqaf, fragStr); 
			fmt.commit();
		}	
		Log.d(LOGTAG, "onCreateView() 99 mSesId:" + mSesId);
		return inflater.inflate(R.layout.fragment_sessiondetail, container, false);
	} // END onCreateView()
	
	/**
	 * As the Activity is created, prepare the LoaderManager and associate screen widgets with Adapters.
	 */
	@SuppressWarnings("deprecation") // Benign complaints about a temporary null cursor used...
	@Override
	public void onActivityCreated(Bundle savedInstanceState) {
		Log.d(LOGTAG, "onActivityCreated() 00   ");
		super.onActivityCreated(savedInstanceState);
	
		// Get handles other widgets and store in member attributes.
		// Ideally these would only be defined here - but the current extraction is a bit clumsy FIXME: polish/redo
		int lcol = 0;
		// RowId
		SessionDetailWidgetMeta wm = new SessionDetailWidgetMeta();
		wm.mDBColPos = lcol++;
		wm.mDBCol = SessionsDB._ID;
		mDetailWidgets.add(wm);
		// DateStamp
		wm = new SessionDetailWidgetMeta();
		wm.mResourceID = R.id.detail_text_datetime;
		wm.mUIWidget = (EditText) mAct.findViewById(wm.mResourceID);
		wm.mWidgetType = SessionDetailWidgetMeta.WT_EDITTEXT;
		wm.mDBColPos = lcol++; 
		wm.mDBCol = SessionsDB.PRACTICE_DATE_STAMP;
		mDetailWidgets.add(wm);
		// length
		wm = new SessionDetailWidgetMeta();
		wm.mResourceID = R.id.detail_text_length;
		wm.mUIWidget = (EditText) mAct.findViewById(wm.mResourceID);
		wm.mWidgetType = SessionDetailWidgetMeta.WT_EDITTEXT;
		wm.mDBColPos = lcol++; 
		wm.mDBCol = SessionsDB.PRACTICE_LENGTH_TOTAL;
		mDetailWidgets.add(wm);
		// noise length
		wm = new SessionDetailWidgetMeta();
		wm.mResourceID = R.id.detail_text_noiselength;
		wm.mUIWidget = (EditText) mAct.findViewById(wm.mResourceID);
		wm.mWidgetType = SessionDetailWidgetMeta.WT_EDITTEXT;
		wm.mDBColPos = lcol++; 
		wm.mDBCol = SessionsDB.PRACTICE_LENGTH_NOISE;
		mDetailWidgets.add(wm);
		// audio filename
		wm = new SessionDetailWidgetMeta();
		wm.mResourceID = R.id.label_audio_file;
		wm.mUIWidget = (TextView) mAct.findViewById(wm.mResourceID);
		wm.mWidgetType = SessionDetailWidgetMeta.WT_TEXTVIEW;
		wm.mDBColPos = lcol++; 
		wm.mDBCol = SessionsDB.PRACTICE_AUDIO_FILE;
		mDetailWidgets.add(wm);		
		// goal
		wm = new SessionDetailWidgetMeta();
		wm.mResourceID = R.id.detail_spinner_goal;
		wm.mUIWidget = (Spinner) mAct.findViewById(wm.mResourceID);
		wm.mWidgetType = SessionDetailWidgetMeta.WT_SPINNER;
		wm.mDBColPos = lcol++; 
		wm.mDBCol = SessionsDB.PRACTICE_GOAL_ID;
		mDetailWidgets.add(wm);
		// notes
		wm = new SessionDetailWidgetMeta();
		wm.mResourceID = R.id.detail_text_notes;
		wm.mUIWidget = (EditText) mAct.findViewById(wm.mResourceID);
		wm.mWidgetType = SessionDetailWidgetMeta.WT_EDITTEXT;
		wm.mDBColPos = lcol++; 
		wm.mDBCol = SessionsDB.PRACTICE_NOTES;
		mDetailWidgets.add(wm);
		// instrument
		wm = new SessionDetailWidgetMeta();
		wm.mResourceID = R.id.detail_spinner_instrument;
		wm.mUIWidget = (Spinner) mAct.findViewById(wm.mResourceID);
		wm.mWidgetType = SessionDetailWidgetMeta.WT_SPINNER;
		wm.mDBColPos = lcol++; 
		wm.mDBCol = SessionsDB.PRACTICE_INSTRUMENT_ID;
		mDetailWidgets.add(wm);
		
		// Initialize adapters to "empty"; they will later be updated by the Loader.
		mAdapterGoals = new SimpleCursorAdapter(mAct,
				android.R.layout.simple_spinner_item, null,
				new String[] { SessionsDB.GOAL_NAME },
				new int[] { android.R.id.text1 });
		mAdapterGoals.setDropDownViewResource(android.R.layout.simple_spinner_dropdown_item);
		mAdapterInstruments = new SimpleCursorAdapter(mAct,
				android.R.layout.simple_spinner_item, null,
				new String[] { SessionsDB.INSTRUMENT_NAME },
				new int[] { android.R.id.text1 });
		mAdapterInstruments.setDropDownViewResource(android.R.layout.simple_spinner_dropdown_item);

		
		// Prepare the Loaders, triggering callback methods. - this should occur after the above widget meta data is populated
		getLoaderManager().initLoader(LDR_INSTRUMENTS, null, this);
		getLoaderManager().initLoader(LDR_GOALS, null, this);
		// connect Spinner controls with CursorAdapter
		Spinner spinnerInstruments = (Spinner) mAct.findViewById(R.id.detail_spinner_instrument);
		spinnerInstruments.setAdapter(mAdapterInstruments);
		Spinner spinnerGoals = (Spinner) mAct.findViewById(R.id.detail_spinner_goal);
		spinnerGoals.setAdapter(mAdapterGoals);
		
		getLoaderManager().initLoader(LDR_SESSIONS, null, this);			

		Log.d(LOGTAG, "onActivityCreated() 99");
	} // END onActivityCreated()

    @Override
    public void onStart() {
    	super.onStart();
    	Log.d(LOGTAG, "onStart() 00");
    }
	
	@Override
	public void onResume() {
		super.onResume();
		Log.d(LOGTAG, "onResume() 00");

		// Update handles to Previous, Next, and Un/Lock buttons for scrolling
		mPrevButton = (Button) mAct.findViewById(R.id.buttonDetailScrollPrev);
		mPrevButton.setOnClickListener(clickerPrev); 
		mNextButton = (Button) mAct.findViewById(R.id.buttonDetailScrollNext);
		mNextButton.setOnClickListener(clickerNext);
		mEditLockButton = (Button) mAct.findViewById(R.id.buttonDetailEditLock);
		mEditLockButton.setOnClickListener(clickerEditLock);
		
		// Update EditLock button text, and other UI widgets enabled state.
		updateSessionDetailWidgets();
		Log.d(LOGTAG, "onResume() 99");
	} // END onResume()

	@Override
	public void onPause() {
		super.onPause();
		Log.d(LOGTAG, "onPause() 00");
	}
	
    @Override
    public void onStop() {
    	Log.d(LOGTAG, "onStop() 00 bound_flag: " + mAudioServiceBound + " this: " + this);		
    	Log.d(LOGTAG, "onStop() 05 mAudioService: " + mAudioService );		
    	super.onStop();
    	if (mAudioServiceBound) {
    		try {
    			mAct.unbindService(mAServiceConnection);
    		} catch (IllegalArgumentException e) {
                Log.e(LOGTAG, "onStop() HACK FIXME: ignoring 'duplicate' fragments trying to un-bind from service.   exception ex: " + e.getLocalizedMessage() );
                e.printStackTrace();
    		}
    	}
//		// Remove sub-Fragment with audio UI.
//		Log.d(LOGTAG, "onDestroyView() 00 audio fragment remove");
//    	String fragStr = getString(R.string.frag_audio_ui);     
//    	FragmentManager fm = getChildFragmentManager();
//		MPQAudioFragment mpqaf = (MPQAudioFragment)  fm.findFragmentByTag(fragStr);
//		if (null != mpqaf) {
//			Log.d(LOGTAG, "onDestroyView() 69 audio fragment found and REMOVING");
//			FragmentTransaction fmt = fm.beginTransaction();
//			fmt.remove(mpqaf); 
//			fmt.commit();
//		}
//		Log.d(LOGTAG, "onDestroyView() 99 audio fragment removed");
    	
    } // END onStop()
	
//    @Override
//    public void onDestroyView() {
//		// Remove sub-Fragment with audio UI.
//		Log.d(LOGTAG, "onDestroyView() 00 audio fragment remove");
//    	String fragStr = getString(R.string.frag_audio_ui);     
//    	FragmentManager fm = getChildFragmentManager();
//		MPQAudioFragment mpqaf = (MPQAudioFragment)  fm.findFragmentByTag(fragStr);
//		if (null != mpqaf) {
//			Log.d(LOGTAG, "onDestroyView() 69 audio fragment found and REMOVING");
//			FragmentTransaction fmt = fm.beginTransaction();
//			fmt.remove(mpqaf); 
//			fmt.commit();
//		}
//		Log.d(LOGTAG, "onDestroyView() 99 audio fragment removed");
//    } // END onDestroyView()
    
	@Override
	public void onDetach() {
		super.onDetach();
		Log.d(LOGTAG, "onDetach() 00");
	}
	
    /**
     * Save state of the app.
     */
    @Override
    public void onSaveInstanceState(Bundle outState) {
    	super.onSaveInstanceState(outState);
    	Log.d(LOGTAG, "onSaveInstanceState() 00");
     	outState.putLong(BNDL_SESSION_ID , mSesId);
    	outState.putInt(BNDL_SESSION_STATE, mSesState);
    	outState.putString(BNDL_SESSION_FILENAME, mFileName);
    	Log.d(LOGTAG, "onSaveInstanceState() 99 outState.size:" + outState.size());
    }	

	
	/**
	 * Listener for Previous button.
	 * {@link #mSesId} is updated with a practice session to open.
	 */
	OnClickListener clickerPrev = new OnClickListener() {
		public void onClick(View v) {	
			Log.d(LOGTAG, "Prev onClick() 00");

			Uri luri = ContentUris.withAppendedId(SessionsDB.CONTENT_PRACTICES_PREV_URI, mSesId);
			Log.d(LOGTAG, "Prev onClick() 10 luri:" + luri);
			Cursor lcursor = mAct.getContentResolver().query(luri, null, null, null, null);
			long lrow = lcursor.getLong(0);
			lcursor.close();
			Log.d(LOGTAG, "Prev onClick() 20 lrow:" + lrow + " *** ");
			
			if (lrow != 0) {
				// Sanity check that mSesId refers to a valid session; alternately get Min.
				int i = 0;
				if (lrow > 0)
					i = getSessionDetailIdCount(lrow);
				if (1 == i) {
			    	mSesId = lrow;
			    } else {
					updateSessionsCount();
					mSesId = mSesMin;
			    }
				mSesActionFlag = FLG_SES_ACTION_OPEN;
				mSesActionId = mSesId;
				updateSessionDetailData();
			} else {
				Toast.makeText(mAct,getString(R.string.button_detail_prev_msg), Toast.LENGTH_SHORT).show();
			}
		}
	}; // END OnClickListener clickerPrev
	/**
	 * Listener for Next button.
	 * {@link #mSesId} is updated with a practice session to open.
	 */
	OnClickListener clickerNext = new OnClickListener() {
		public void onClick(View v) {
			Log.d(LOGTAG, "Next onClick() 00");
			
			Uri luri = ContentUris.withAppendedId(SessionsDB.CONTENT_PRACTICES_NEXT_URI, mSesId);
			Log.d(LOGTAG, "Next onClick() 10 luri:" + luri);
			Cursor lcursor = mAct.getContentResolver().query(luri, null, null, null, null);
			long lrow = lcursor.getLong(0);
			lcursor.close();
			Log.d(LOGTAG, "Next onClick() 20 lrow:" + lrow + " *** ");
			
			if (lrow != 0) {
				// Sanity check that mSesId refers to a valid session; alternately get Min.
				int i = 0;
				if (lrow > 0)
					i = getSessionDetailIdCount(lrow);
				if (1 == i) {
			    	mSesId = lrow;
			    } else {
					updateSessionsCount();
					mSesId = mSesMax;
			    }
				mSesActionFlag = FLG_SES_ACTION_OPEN;
				mSesActionId = mSesId;
				updateSessionDetailData();
			} else {
				Toast.makeText(mAct,getString(R.string.button_detail_next_msg), Toast.LENGTH_SHORT).show();				
			}
		}
	}; // END OnClickListener clickerNext
	
	/**
	 * Respond to edit/unlock button event.
	 *   Change state for the current session detail.
	 *     Set a flag to save data in a bundle when exiting screen mSesState.
	 *     Lock/Unlock the input fields.
	 *     Update button icon to unlock or save/cancel.
	 */
	OnClickListener clickerEditLock = new OnClickListener() {
		public void onClick(View v) {
			Log.d(LOGTAG, "EditLock onClick() 00");

			// toggle state and update EditLock button text, based on state
			if (FLG_SES_STATE_VIEW == mSesState) {
				mSesState = FLG_SES_STATE_EDIT;
				// Update EditLock button text and UI widgets.
				updateSessionDetailWidgets();			
			} else if (FLG_SES_STATE_EDIT == mSesState){
				// Prompt user if they want to save or cancel changes.
            	MPQGenDialogFragment ldlg = new MPQGenDialogFragment();
            	Bundle args = new Bundle();
            	args.putInt(MPQGenDialogFragment.ARG_DIALOG_TYPE, MPQGenDialogFragment.ARG_DIALOG_TYPE_SAVE_DISCARD);
            	args.putString(MPQGenDialogFragment.ARG_DIALOG_TITLE, getString(R.string.dialog_mpqadd_title_save_discard));
            	args.putInt(MPQGenDialogFragment.ARG_DIALOG_ID, DLG_SAVE_DISCARD);
            	ldlg.setArguments(args);
            	ldlg.setTargetFragment( mAct.getSupportFragmentManager().findFragmentByTag(getString(R.string.frag_tab2_detail))  , MPQGenDialogFragment.ARG_DIALOG_TYPE_SAVE_DISCARD);
            	ldlg.show(mAct.getSupportFragmentManager(), getString(R.string.frag_dialog_save_discard));
			}
		}
	}; // END OnClickListener clickerEditLock

	
	/**
	 * MPQGenDialog Positive Click Callback.
	 * @param dialog DialogFragment
	 * @param id int to identify callback
	 * @throws IllegalArgumentException for unknown id
	 */
	public void onMPQGenDialogPositiveClick(DialogFragment dialogFrag, int id) {
		Log.d(LOGTAG, "onMPQGenDialogPOS() 00");
		
	    // Call ContentResolver with uri and save values from dialog.
		if (id == DLG_SAVE_DISCARD) {			
			Log.d(LOGTAG, "Detail CallBack onMPQGenDialogPositiveClick() 10 id SAVE DISCARD");
			// FIXME: very warty - very little data abstraction -- a detail session object is likely called for... for now, deliver something
			// FIXME: only run this/update if data has been edited by the user (should keep a[nother] state flag) ... not too easy to implement though, as it means every control needs to be watched, i believe; alternately, get DB data, and compare widgets with DB data... (still a good deal of effort)
			
			// Collect data from controls to store
			ContentValues newValues = new ContentValues();
			SessionDetailWidgetMeta lwm;
			long ltimeSec;
			String ltext;
			int lrowCount;	// Count of how many rows were updated (should be 1).
			long lrowId;
			int lcol = 1;
			EditText let; // Local EditText view handle
			Spinner lsp;  // Local Spinner handle;
			Uri luri;
			// DateStamp
			lwm = mDetailWidgets.get(lcol++); // FIXME: horrible way to loop through controls
			let = (EditText) lwm.mUIWidget;
			ltext = let.getText().toString(); 	
			ltimeSec = SessionsDB.strDateLong(ltext);
			newValues.put(SessionsDB.PRACTICE_DATE_STAMP, ltimeSec);
			// length
			lwm = mDetailWidgets.get(lcol++);
			let = (EditText) lwm.mUIWidget;
			ltext = let.getText().toString(); 	
			ltimeSec = SessionsDB.strSessionMSecLong(ltext);
			newValues.put(SessionsDB.PRACTICE_LENGTH_TOTAL, ltimeSec);
			// noise length
			lwm = mDetailWidgets.get(lcol++);
			let = (EditText) lwm.mUIWidget;
			ltext = let.getText().toString(); 	
			ltimeSec = SessionsDB.strSessionMSecLong(ltext);
			newValues.put(SessionsDB.PRACTICE_LENGTH_NOISE, ltimeSec);
			// audio filename (skipped over; audio file is fixed to session)
			lcol++; // still need to increment list index 
			// goal 
			lwm = mDetailWidgets.get(lcol++);
			lsp = (Spinner) lwm.mUIWidget;
			lrowId = lsp.getSelectedItemId(); 
			newValues.put(SessionsDB.PRACTICE_GOAL_ID, lrowId);
			// notes
			lwm = mDetailWidgets.get(lcol++); // FIXME: horrible way to loop through
			let = (EditText) lwm.mUIWidget;
			ltext = let.getText().toString(); 	
			newValues.put(SessionsDB.PRACTICE_NOTES, ltext);
			// instrument
			lwm = mDetailWidgets.get(lcol++);
			lsp = (Spinner) lwm.mUIWidget;
			lrowId = lsp.getSelectedItemId(); 
			newValues.put(SessionsDB.PRACTICE_INSTRUMENT_ID, lrowId);

			luri = ContentUris.withAppendedId(SessionsDB.CONTENT_PRACTICES_URI, mSesId);
			lrowCount = mAct.getContentResolver().update(luri, newValues, null, null);
			Log.d(LOGTAG, "Detail CallBack onMPQGenDialogPositiveClick() 18 newValues:" + newValues.toString() );
			Log.d(LOGTAG, "Detail CallBack onMPQGenDialogPositiveClick() 19  rowCount:" + lrowCount );

			// Toggle state and update locked widgets
			mSesState = FLG_SES_STATE_VIEW;
			updateSessionDetailWidgets();
			updateSessionDetailData(); 
		} else {
            throw new IllegalArgumentException("SessionDetailFragment onMPQGenDialogPositiveClick() Unknown DLG_ id: " + id);
		}		
		dialogFrag.dismiss();
	} // END onMPQGenDialogPositiveClick()
	
	/**
	 * MPQGenDialog Negative Click Callback.  Really nothing to do, other than dismiss the dialog.
	 * @param dialogFrag DialogFragment
	 * @param id int to identify callback
	 * @throws IllegalArgumentException for unknown id
	 */
	public void onMPQGenDialogNegativeClick(DialogFragment dialogFrag, int id) {
		Log.d(LOGTAG, "onMPQGenDialogNeg() 00 id: " + id);
		if (id == DLG_SAVE_DISCARD) {
			// Toggle state and update locked widgets
			mSesState = FLG_SES_STATE_VIEW;
			updateSessionDetailWidgets();
			updateSessionDetailData(); // FIXME: only run this if data has been edited by the user (should keep a[nother] state flag) ... not too easy to implement though, as it means every control needs to be watched, i believe
		} else {
            throw new IllegalArgumentException("SessionDetailFragment onMPQGenDialogPositiveClick() Unknown DLG_ id: " + id);
		}
		dialogFrag.dismiss();
	}	
	
	// ///////////////////////////////////////////////////
	// LoaderManager.LoaderCallbacks<Cursor> implementations
	/**
	 * LoaderManager callback to create new loaders.
	 * {@link #mSesId} is read with the practice session to load.
	 */
	public Loader<Cursor> onCreateLoader(int id, Bundle args) {
		Log.d(LOGTAG, "onCreateLoader() 00 id: " + id);

		String[] lproj;
		String lselc = null;
		String[] lsela = { "" };
		Uri luri;
		
		// Setup projection		
		if (LDR_INSTRUMENTS == id) {
			Log.d(LOGTAG, "onCreateLoader() 10 id: " + id);
			lproj = new String[] { SessionsDB._ID, SessionsDB.INSTRUMENT_NAME };
			luri = SessionsDB.CONTENT_INSTRUMENTS_URI;
		} else if (LDR_GOALS == id) {
			Log.d(LOGTAG, "onCreateLoader() 12 id: " + id);
			lproj = new String[] { SessionsDB._ID, SessionsDB.GOAL_NAME };
			luri = SessionsDB.CONTENT_GOALS_URI;
		} else if (LDR_SESSIONS == id) {
			Log.d(LOGTAG, "onCreateLoader() 14 id: " + id);			
			SessionDetailWidgetMeta lwm;
			lproj = new String[mDetailWidgets.size()]; // temporary array for projection, the size of the number of columns we'll be getting
			for (int i=0; i < mDetailWidgets.size(); i++) { // FIXME; want to ensure we're in order of mDBColPos
				lwm = mDetailWidgets.get(i);
				lproj[i] = lwm.mDBCol;
				Log.d(LOGTAG, "onCreateLoader() 15a lproj[" + i + "]: " + lproj[i]);
			}
			luri = SessionsDB.CONTENT_PRACTICES_URI;
			luri = ContentUris.withAppendedId(luri, mSesId);
			
			// Update max row count, and store it in a member attribute.
			updateSessionsCount();
		} else {
			Log.d(LOGTAG, "onCreateLoader() 16 id: " + id);			
            throw new IllegalArgumentException("onCreateLoader() id: " + id);
		}

		// Get cursor
		CursorLoader lcursLoader = new CursorLoader(mAct,
				luri, lproj, lselc, lsela, null);
		
		Log.d(LOGTAG, "onCreateLoader() 99");
		return lcursLoader;
	} // END onCreateLoader()

	/**
	 * LoaderManager callback to process when loading cursor data is completed.
	 */
	public void onLoadFinished(Loader<Cursor> loader, Cursor data) {
		int id = loader.getId();
		Log.d(LOGTAG, "onLoadFinished() 00 id: " + id);
		if (null == data) {
			Log.d(LOGTAG, "onLoadFinished() 02 - data is null - return now");
			return;
		}
		if (LDR_INSTRUMENTS == id) {
			int lid;
			Spinner lsp = (Spinner) mAct.findViewById(R.id.detail_spinner_instrument);
 			// Be careful not to overwrite selected item in the spinner.
			lid = lsp.getSelectedItemPosition();
			Log.d(LOGTAG, "onLoadFinished() 02 Instrument data lid: " + lid );
			mAdapterInstruments.swapCursor(data);
			lsp.setSelection(lid); 
		} else if (LDR_GOALS == id) {
			int lid;
			Spinner lsp = (Spinner) mAct.findViewById(R.id.detail_spinner_goal);
 			// Be careful not to overwrite selected item in the spinner.
			lid = lsp.getSelectedItemPosition();
			mAdapterGoals.swapCursor(data);
			lsp.setSelection(lid); 
		} else if (LDR_SESSIONS == id) {
			Log.d(LOGTAG, "onLoadFinished() 10 Session data columns: " + Arrays.toString(data.getColumnNames()) );
			Log.d(LOGTAG, "onLoadFinished() 12 Session data pos: " + data.getPosition() + " count: " + data.getCount() + " columnCount: " + data.getColumnCount() + " mSesId:" + mSesId);
			// FIXME: very warty - very little data abstraction -- a detail session object is likely called for... for now, deliver something
			// Populate controls, ideally these would be looped through consistently -- blocking is the variety in data/widget types...
			SessionDetailWidgetMeta lwm;
			long ltimeSec;
			String ltext;
			int lid;
			int lcol = 1;
			EditText let; // local EditText view handle
			TextView ltv; // local TextView handle;
			Spinner lsp;  // local Spinner handle;
			// DateStamp
			lwm = mDetailWidgets.get(lcol++); // FIXME: horrible way to loop through controls
			ltimeSec = data.getLong(lwm.mDBColPos);
			let = (EditText) lwm.mUIWidget;
			let.setText(SessionsDB.longDateString(ltimeSec)); 		
			// length
			lwm = mDetailWidgets.get(lcol++);
			ltimeSec = data.getLong(lwm.mDBColPos);
			let = (EditText) lwm.mUIWidget;
			let.setText( SessionsDB.longSessionMSecString(ltimeSec));		
			// noise length
			lwm = mDetailWidgets.get(lcol++);
			ltimeSec = data.getLong(lwm.mDBColPos);
			let = (EditText) lwm.mUIWidget;
			let.setText(SessionsDB.longSessionMSecString(ltimeSec));					
			// audio filename 
			lwm = mDetailWidgets.get(lcol++);
			ltext = data.getString(lwm.mDBColPos);
			Log.d(LOGTAG, "onLoadFinished() 14 filename:" + ltext + ":");
		  updateFileName(ltext);
			ltv = (TextView) lwm.mUIWidget;
			ltv.setText("audio file: " + ltext + "\n  sessions _ID:" + data.getLong(0) + " min/Max|count:" + mSesMin + "/" + mSesMax + "|" + mSesCount);
			lwm = mDetailWidgets.get(lcol++);
			lid = data.getInt(lwm.mDBColPos);
			Log.d(LOGTAG, "onLoadFinished() 16 Session DB lid: " + lid + " col: " + lwm.mDBCol );
			if (lid > 0) lid--; // Translate row id into spinner id; former is indexed on 1, latter on 0.
			lsp = (Spinner) lwm.mUIWidget;
			lsp.setPrompt("Goals:");
			lsp.setSelection(lid); 
			// notes
			lwm = mDetailWidgets.get(lcol++); // FIXME: horrible way to loop through
			ltext = data.getString(lwm.mDBColPos);
			let = (EditText) lwm.mUIWidget;
			let.setText(ltext);								
			// instrument
			lwm = mDetailWidgets.get(lcol++);
			lid = data.getInt(lwm.mDBColPos);
			Log.d(LOGTAG, "onLoadFinished() 18 Session DB lid: " + lid + " col: " + lwm.mDBCol );
			if (lid > 0) lid--; // translate row id into spinner id; former is indexed on 1, latter on 0
			lsp = (Spinner) lwm.mUIWidget;
			lsp.setPrompt("Instruments:");
			lsp.setSelection(lid);
			
			// Reset View/Edit state, and update UI widgets.
			mSesState = FLG_SES_STATE_VIEW;
			updateSessionDetailWidgets();
		} else {
            throw new IllegalArgumentException("onFinishedLoader() getId: " + id);
		}
		Log.d(LOGTAG, "onLoadFinished() 99 id: " + id);
	} // END onLoadFInished()

	/**
	 * LoaderManager callback when releasing data.
	 */
	public void onLoaderReset(Loader<Cursor> loader) {
		int id = loader.getId();
		Log.d(LOGTAG, "onLoaderReset() 00");
		if (LDR_INSTRUMENTS == id) {
			mAdapterInstruments.swapCursor(null);
		} else if (LDR_GOALS == id) {
			mAdapterGoals.swapCursor(null);
		}  else if (LDR_SESSIONS == id) {
			; // pass for now - data was copied into static widgets - no persistent cursor to maintain.  - should data be written back to disk here? does not feel appropriate...
		} else  {
            throw new IllegalArgumentException("onLoaderReset() getId: " + id);
		}
	} // END onLoaderReset()

	
	/*****************************************
	 * Private helper bind-connection callbacks and Listeners.  
	 */
	
    /**
     * Callbacks for service binding, passed to bindService().
     */
    private ServiceConnection mAServiceConnection = new ServiceConnection() {
    	public void onServiceConnected(ComponentName className, IBinder service_binder) {
    		MPQAudioService.LocalBinder binder = (MPQAudioService.LocalBinder) service_binder;
    		mAudioService = binder.getService();
    		mAudioService.setListenerBnd(aserviceListener, getTag());
    		mAudioServiceBound = true;

    		// FIXME: check service state and react? -- mAudioServiceState = mAudioService.getStateBnd();
			Log.w(LOGTAG, "mAServiceConnection onServiceConnected() 99 got State: " + mAudioService.getStateBnd());
    	} // END onServiceConnected()
    	public void onServiceDisconnected(ComponentName className) {
    		if (null != mAudioService) {
    			mAudioService.delListenerBnd(getTag());
    		}
    		mAudioServiceBound = false;
			Log.w(LOGTAG, "mAServiceConnection onServiceDisconnected() 99 classname: " + className);
    	}
    }; // END ServiceConnection mAServiceConnection
    
    
    
	/**
	 * MPQ Audio Service Listener.
	 */
	private MPQAudioService.MPQAudioServiceListener aserviceListener = new MPQAudioService.MPQAudioServiceListener() {
		public void onMPQAudioServiceStateChange(int newState) {
			Log.w(LOGTAG, "aserviceListener  onMPQAudioServiceStateChange() 00  new State: " + newState);
			// Trigger UI updates.	
			updateSessionDetailWidgets();
		} // END onMPQAudioServiceStateChange();
		
		public void onMPQAudioServiceAudioChange(long newPos, long fileLen, int newVol) {
			Log.w(LOGTAG, "aserviceListener  onMPQAudioServiceAudioChange() 00  new Pos: " + newPos + " new Vol: " + newVol);
			// for now - do nothing other than log the call.
		} // END onMPQAudioServiceAudioChange()
		
		public void onMPQAudioServiceAADone(String fileName, long audioLenMS, long noiseLenMS) {
			Log.w(LOGTAG, "aserviceListener  onMPQAudioServiceAADone() 00 COMMENTED OUT FOR NOW *** ONE LAYER LOWER *** filename: " + fileName + " audio: " + audioLenMS + " noise: " + noiseLenMS + " mFileName: " + mFileName + " mSesId: " + mSesId);
//			long laaSesId = 0;
//			Uri luri;
//
//			// Get correct session id for the audio analysis.
//			if (fileName.equals(mFileName)) {
//				laaSesId = mSesId;
//			} else {
//				luri = SessionsDB.CONTENT_PRACTICES_FILE_URI;
//				String[] lselectionArgs = new String[] { fileName };
//				Log.w(LOGTAG, "aserviceListener  onMPQAudioServiceAADone() 27 luri:" + luri + " lselArgs: " + lselectionArgs.toString());
//				Cursor lcursor = mAct.getContentResolver().query(luri, null, null, lselectionArgs, null);
//				if (1 == lcursor.getCount()) {
//					laaSesId = lcursor.getLong(0);
//					Log.w(LOGTAG, "aserviceListener  onMPQAudioServiceAADone() 28 laaSesId:" + laaSesId + " msesid: " + mSesId);
//				}
//				lcursor.close();
//			}			
//
//			// update DB with new values.
//			ContentValues newValues = new ContentValues();			
//			newValues.put(SessionsDB.PRACTICE_LENGTH_TOTAL, audioLenMS);
//			newValues.put(SessionsDB.PRACTICE_LENGTH_NOISE, noiseLenMS);		
//			Log.d(LOGTAG, "aserviceListener  onMPQAudioServiceAADone() 70 aasesid: " + laaSesId + " newValues:" + newValues.toString() );
//			luri = ContentUris.withAppendedId(SessionsDB.CONTENT_PRACTICES_URI, laaSesId);
//			int lrowCount = mAct.getContentResolver().update(luri, newValues, null, null);
//			
			Log.w(LOGTAG, "aserviceListener  onMPQAudioServiceAADone() 99 "); // update rowcount:" + lrowCount );
		} // END onMPQAudioServiceAADone()
	}; // END MPQAudioService.MPQAudioServiceListener aserviceListener

    
	

	/**
	 * Utility method to correctly update the member attribute storing the filename, 
	 * and associate filename with AudioService.
	 *   Checks for a running AudioService.  If service is found in REC or PLAY states,
	 *   service trumps the updateFileName request.
	 * @param fn String representing the path to the file on the SD device. ex: "/mpq/file01.3gp"
	 * {@link #mSesActionFlag} and {@link #mSesActionId} are checked for code to OPEN or REC a session. 
	 * {@link #mSesId} is updated if a running AudioService is found, along with the mSesAction variables. 
	 */
	private void updateFileName(String fn) {
		Log.v(LOGTAG, "updateFileName() 00  new fn: " + fn + " mFilename: " + mFileName);

		// Check if there is an already running service, which trumps the update filename request with an early return.
		Log.d(LOGTAG, "updateFileName() 20 AudioService check ... bound:" + mAudioServiceBound);
		if ( (null != mAudioService) &&  (true == mAudioServiceBound) ) {
			int lservice_state = mAudioService.getStateBnd();
			String lservice_filename = mAudioService.getFilenameBnd();
			Log.d(LOGTAG, "updateFileName() 25 AudioService state:" + lservice_state + " file: " + lservice_filename);

			// Get corresponding session ID if service found to be running.
			if ( (MPQAudioService.FLG_SRV_STATE_PLAY == lservice_state) ||
    				(MPQAudioService.FLG_SRV_STATE_PAUSE_PLAY == lservice_state) ||
    				(MPQAudioService.FLG_SRV_STATE_REC == lservice_state) ||
    				(MPQAudioService.FLG_SRV_STATE_PAUSE_REC == lservice_state) ) {
				long lservice_sesid = 0;
				Log.d(LOGTAG, "updateFileName() 26  found running service   old mfile:" + mFileName + " service file: " + lservice_filename);
				Uri luri = SessionsDB.CONTENT_PRACTICES_FILE_URI;
				String[] lselectionArgs = new String[] { lservice_filename };
				Log.d(LOGTAG, "updateFileName() 27 luri:" + luri + " lselArgs: " + lselectionArgs.toString());
				Cursor lcursor = mAct.getContentResolver().query(luri, null, null, lselectionArgs, null);
				if (1 == lcursor.getCount()) {
					lservice_sesid = lcursor.getLong(0);
					Log.d(LOGTAG, "updateFileName() 28 lservice_sesid:" + lservice_sesid + " msesid: " + mSesId + " mSesActionId: " + mSesActionId);

					// Unwind REC command, if we find the audio service already working on something else.
					if ( (FLG_SES_ACTION_REC == mSesActionFlag) && (mSesActionId != lservice_sesid) &&
							(lservice_sesid > 0) ) {
				        Uri lsessionUri = ContentUris.withAppendedId(SessionsDB.CONTENT_PRACTICES_URI, mSesActionId);
			        	Log.d(LOGTAG, "updateFileName() 28A   DELETE action_sesid: " + mSesActionId + " flag: " + mSesActionFlag);
			        	mAct.getContentResolver().delete(lsessionUri, null, null);
			        	Log.d(LOGTAG, "updateFileName() 28B   delete finished");
					} // END if Unwind REC
				}
				lcursor.close();
		
				// Found a valid service session id.
				if (lservice_sesid > 0) {
					// If service is working with a different file, then drop current open/rec action, update mSesId, updateSessionDetailData, and exit.
					if (lservice_sesid != mSesId) {
						mSesId = lservice_sesid;
						updateSessionDetailData();
			        	Log.d(LOGTAG, "updateFileName() 28C 999   EXIT after finding mismatch running service.");						
					} else { // else service is working with same file, then drop open/rec action.
			        	Log.d(LOGTAG, "updateFileName() 28D found service matching session id EXIT.");
					}
					// Reset action state variables.
					mFileName = lservice_filename;
					mSesActionFlag = FLG_SES_ACTION_OPEN;
					mSesActionId = 0;
					return;
				} else { // Service has an invalid session id; stop whatever it's doing.  Proceed with update filename actions below.
					Intent intent = new Intent(mAct, MPQAudioService.class);
					intent.setAction(MPQAudioService.ACTION_STOP);
					mAct.startService(intent);
					Log.d(LOGTAG, "updateFileName() 28E STOP Intent sent");
				}
			} // END if service found running.
		} else Log.w(LOGTAG, "updateFileName() 29   null AudioService ...  probably not running previously??? ***");  // FIXME: need to stall, and try again -- possibly use a delayed Handler ...
			
		// Sanity checks on parameter.
		String lfilename = null;
		if ( (null == fn) || (0 == fn.length()) ) {
			lfilename = null;
		} else {
			if (fn.startsWith(File.separator)) {
				lfilename = fn;
			} else {
				lfilename = File.separator + fn;				
			}
		}  // END if (null == fn)	
		
		if ( (lfilename != null) && !(lfilename.equals(mFileName)) ) {
			Log.d(LOGTAG, "updateFileName() 70 mFileName:" + mFileName + ":being set to lfilename:" + lfilename + ":");
			mFileName = lfilename;

			// Check Action (REC/OPEN).
			if ( (mSesActionId > 0) && (mSesId == mSesActionId) ) {
				// Send Intent to the service of the filename change.
				Intent intent = new Intent(mAct, MPQAudioService.class);

				if (FLG_SES_ACTION_OPEN == mSesActionFlag) {
					intent.setAction(MPQAudioService.ACTION_OPEN);				
				} else if (FLG_SES_ACTION_REC == mSesActionFlag) {
					intent.setAction(MPQAudioService.ACTION_REC);										
				} else {
					Log.e(LOGTAG, "updateFileName() 73   unexpected mSesActionFlag *** " + mSesActionFlag);
				}
				
				// Pass corresponding filename via "extra-bundle".  
		    	Bundle lextraArgs = new Bundle();
		    	lextraArgs.putString(MPQAudioService.ARG_FILENAME, mFileName);
		    	intent.putExtras(lextraArgs);
				mAct.startService(intent);
				Log.d(LOGTAG, "updateFileName() 74 Intent sent filename: " + lfilename + " action id: " + mSesActionId + " action flag: " + mSesActionFlag + " mSesId: " + mSesId);
				// Reset action state variables.
				mSesActionFlag = FLG_SES_ACTION_OPEN;
				mSesActionId = 0;
			} else {
				Log.e(LOGTAG, "updateFileName() 79   mismatch mSesId/mSesActionId *** " + mSesId + "/" + mSesActionId);
			}
		}
		Log.d(LOGTAG, "updateFileName() 99 fn:" + fn + ": mFileName:" + mFileName);
	} // END updateFileName(String)
	
	/**
	 * Utility method to re-query the Practice Sessions table.  This is used when browsing through records.
	 */
	private void updateSessionDetailData() {
		getLoaderManager().restartLoader(LDR_SESSIONS, null, this);
	}
	
	/**
	 * Utility method to loop through UI widgets, and enable or disable them 
	 * based on session state.  Audio Service state is also checked.
	 * {@link #mSesState} is read for session sate
	 * {@link #mDetailWidgets} is read for widgets to update
	 */
	private void updateSessionDetailWidgets() {
		Log.d(LOGTAG, "updateSessionDetailWidgets() 00 state: " + mSesState);
		
		SessionDetailWidgetMeta lwm; // Local widget meta object.
		View lev; 	// Local View handle.
		
		// Exit without updating UI, if the fragment is not attached to an activity.
		Activity la = getActivity();
		if (null == la) {
			Log.e(LOGTAG, "updateSessionDetailWidgets() 05->99  EXIT , called with no current activity la:" + la + " mAct: " + mAct );
			return;
		}

		//loop through widgets and lock/unlock
		for (int i = 0; i < mDetailWidgets.size(); i++ ) {
			lwm = mDetailWidgets.get(i);
			if ( (SessionDetailWidgetMeta.WT_EDITTEXT == lwm.mWidgetType) ||
					(SessionDetailWidgetMeta.WT_SPINNER == lwm.mWidgetType) ) {
				lev = (View) lwm.mUIWidget;
				// Toggle enabled state appropriately.
				if (FLG_SES_STATE_VIEW == mSesState) {
					lev.setEnabled(false);
				} else {
					lev.setEnabled(true);
				}
			}
		} // END for widget loop
		
		// Toggle button state and text.
		if (FLG_SES_STATE_VIEW == mSesState) {
			mEditLockButton.setText(getString(R.string.button_detail_unlock));
			mPrevButton.setEnabled(true);
			mNextButton.setEnabled(true);
			mAct.mTabHost.getTabWidget().setEnabled(true);
		} else {
			mEditLockButton.setText(getString(R.string.button_detail_save)); // FIXME: probably need a better string identifier name
			mPrevButton.setEnabled(false);
			mNextButton.setEnabled(false);
			mAct.mTabHost.getTabWidget().setEnabled(false);
		}

		// If audio service state is playing, lock navigation controls.
		Log.d(LOGTAG, "updateSessionDetailWidgets() 20 AudioService check ... bound:" + mAudioServiceBound);
		if ( (null != mAudioService) &&  (true == mAudioServiceBound) ) {
			int lservice_state = mAudioService.getStateBnd();
			Log.d(LOGTAG, "updateSessionDetailWidgets() 25 AudioService state:" + lservice_state);
			if ( (MPQAudioService.FLG_SRV_STATE_PLAY == lservice_state) ||
    				(MPQAudioService.FLG_SRV_STATE_PAUSE_PLAY == lservice_state) ||
    				(MPQAudioService.FLG_SRV_STATE_REC == lservice_state) ||
    				(MPQAudioService.FLG_SRV_STATE_PAUSE_REC == lservice_state) ||
    				(MPQAudioService.FLG_SRV_STATE_STOP == lservice_state) ) {
				mPrevButton.setEnabled(false);
				mNextButton.setEnabled(false);
			}
		} else Log.w(LOGTAG, "updateSessionDetailWidgets() 29   null AudioService ...  probably not running previously??? ***");  // FIXME: need to stall, and try again -- possibly use a delayed Handler ...
		

// FIXME: dig into this -- it's not working :(		
//		// Disable the keyboard if state is to view.
//		if (FLG_SES_STATE_VIEW == mSesState) {
//			InputMethodManager imm = (InputMethodManager) getActivity().getSystemService(Context.INPUT_METHOD_SERVICE);
//			imm.hideSoftInputFromWindow(getView().getWindowToken(), 0);
//		}
	} // END updateSessionDetailWidgets()
	
	/**
	 * Helper method to query the DB and update the min and max values and store them in member attributes.
	 */
	private void updateSessionsCount() {
		Cursor lcur = mAct.getContentResolver().query(SessionsDB.CONTENT_PRACTICES_COUNT_URI, null, null, null, null);
		Log.d(LOGTAG, "updateSessionsCount() 00   old mSesMin: " + mSesMin + " mSesMax: " + mSesMax + " mSesCount: " + mSesCount);
		mSesCount = lcur.getLong(0);
		mSesMin = lcur.getLong(1);
		mSesMax = lcur.getLong(2);
		Log.d(LOGTAG, "updateSessionsCount() 90   NEW mSesMin: " + mSesMin + " mSesMax: " + mSesMax + " mSesCount: " + mSesCount);
		lcur.close();
	}

	/**
	 * Helper method to query the DB for a session.  
	 * Returns number of rows matching the session id - 1 or 0.
	 */
	private int getSessionDetailIdCount(long rowId) {
		int li = 0;
		Uri luri = ContentUris.withAppendedId(SessionsDB.CONTENT_PRACTICES_URI, rowId);
		Cursor lcur = mAct.getContentResolver().query(luri, null, null, null, null);
		if (null != lcur) {
			li = lcur.getCount();
			lcur.close();		
		}
		Log.d(LOGTAG, "getSessionDetailIdCount() 15  count(): " + li );
		return li;
	}
	
	
	/**
	 * Utility method to create a new practice session.
	 * @returns long with Row ID of the new row
	 */
	private long newSessionDetailData() {
		Log.d(LOGTAG, "newSessionDetailData() 00  ");
		
		SimpleDateFormat fileDateFormat = new SimpleDateFormat("yyyyMMdd_HHmm"); // FIXME: this should likely be moved to strings.xml
		long lNowSec = java.lang.System.currentTimeMillis();
		
		// Calculate filename, based on date-stamp.
		SharedPreferences sharedPref = PreferenceManager.getDefaultSharedPreferences(mAct);
		String lDirSD = sharedPref.getString(getString(R.string.pref_audiodirectory_key), getString(R.string.pref_audiodirectory_default));
		
		// FIXME: should add more checks: available disk space check, consumed disk space check, total size used check
		String lFileName = lDirSD + "session_";		
		lFileName = lFileName + fileDateFormat.format(new Date(lNowSec)) + ".wav";
		Log.d(LOGTAG, "newSessionDetailData() 20  filename: " + lFileName);
		
		// Insert new session - with filename into DB.		
        ContentValues val = new ContentValues();
        val.put(SessionsDB.PRACTICE_DATE_STAMP, lNowSec);
        val.put(SessionsDB.PRACTICE_AUDIO_FILE, lFileName);
        Uri lNewUri;
		lNewUri = mAct.getContentResolver().insert(SessionsDB.CONTENT_PRACTICES_URI, val);
		long lNewRow = ContentUris.parseId(lNewUri);

		Log.d(LOGTAG, "newSessionDetailData() 99 lNewRow: " + lNewRow);
		return lNewRow;
	} // END newSessionDetailData()
	
	
} // END class SessionDetailFragment
