/**
 * SessionsListFragment.java: List of sessions.
 *            Copyright (C) 2012 donfede@casagrau.org
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * created 2012-09-23 donfede@casagrau.org
 * credit to: package com.example.android.notepad/NotesList.java
 *
 * $Id: SessionsListFragment.java,v 1.28 2013-03-31 21:18:03 donfede Exp $
 */
package org.casagrau.mpq.mpq_android;

import org.casagrau.mpq.audio_service.MPQAudioService;
import org.casagrau.mpq.mpq_android.R.layout;
import org.casagrau.mpq.mpq_android.SessionsDBProvider.SessionsDB;

import android.app.Activity;
import android.content.ContentUris;
import android.content.Intent;
import android.database.Cursor;
import android.net.Uri;
import android.os.Bundle;
import android.support.v4.app.Fragment;
import android.support.v4.app.LoaderManager;
import android.support.v4.content.CursorLoader;
import android.support.v4.content.Loader;
import android.support.v4.widget.SimpleCursorAdapter;
import android.support.v4.widget.SimpleCursorAdapter.ViewBinder;
import android.util.Log;
import android.view.ContextMenu;
import android.view.ContextMenu.ContextMenuInfo;
import android.view.LayoutInflater;
import android.view.MenuInflater;
import android.view.MenuItem;
import android.view.View;
import android.view.ViewGroup;
import android.widget.AdapterView;
import android.widget.ListView;
import android.widget.TextView;

/**
 * Fragment with GUI for Sessions List.
 */
public class SessionsListFragment extends Fragment implements
		ListView.OnItemClickListener, LoaderManager.LoaderCallbacks<Cursor> {
	private static final String LOGTAG = "SessionsListFragment";
	
	private MPQMainTabActivity mAct;	// Local caching variable with the activity.

	// Unique integer identifiers for Loaders.
	private static final int LDR_SESSIONS = 2;
	private static final int LDR_SESSIONS_TOT = 3;

	// Adapters to bind to CursorLoaders
	private SimpleCursorAdapter mAdapterSessions;

	// Handles to various UI widgets.
	private TextView mLabelTotLen;
	private TextView mLabelTotNoise;
	
	// Parameterless constructor, required by framework
	public SessionsListFragment() {
		super();
	}
	
	@Override
	public void onAttach(Activity a) {
		super.onAttach(a);
		Log.d(LOGTAG, "onAttach() 00");
		mAct = (MPQMainTabActivity) a;
	}
		
	/**
	 * Create the view for this fragment - bringing XML layout with list that has customized columns.
	 */
	@Override
	public View onCreateView(LayoutInflater inflater, ViewGroup container, 
			Bundle savedInstanceState) {	
		View v = inflater.inflate(R.layout.fragment_sessionslist, container, false);
		return v;
	}
	
	
	@SuppressWarnings("deprecation") // benign complaints about a temporary null cursor used...
	@Override
	public void onResume() {
		super.onResume();

		// bind adapter to ListView and connect listener
		String[] fromColumns = {SessionsDB.PRACTICE_DATE_STAMP, 
				SessionsDB.PRACTICE_LENGTH_TOTAL, 
				SessionsDB.PRACTICE_LENGTH_NOISE };
		int[] toViews = {R.id.list_date_stamp, 
				R.id.list_length_total, 
				R.id.list_length_noise};
		mAdapterSessions = new SimpleCursorAdapter(mAct,
				layout.fragment_sessionslist_item, null,
				fromColumns, toViews );

		// Customize the output of the list layout.
		mAdapterSessions.setViewBinder(new ViewBinder() {
			public boolean setViewValue(View aView, Cursor aCursor,
					int col) {
				long lMSec;
				String lStr;
				// FIXME: a lot of cleanup/tidy that could be done in this if() statement ... for now - press on.
				if (1 == col) { 
					lMSec = aCursor.getLong(col);
			        lStr = SessionsDB.longDateString(lMSec);
					TextView textView = (TextView) aView;
					textView.setText(lStr);
					return true;
				} else if (2 == col) {
					lMSec = aCursor.getLong(col);
					lStr = "Len: " + SessionsDB.longSessionMSecString(lMSec);
					TextView textView = (TextView) aView;
					textView.setText(lStr);
					return true;
				} else if (3 == col) {
					lMSec = aCursor.getLong(col);
					lStr =  "Noise: " + SessionsDB.longSessionMSecString(lMSec);
					TextView textView = (TextView) aView;
					textView.setText(lStr);
					return true;
				}
				return false;
			}
		}); // END setViewBinder() 

		// Prepare the Loaders, triggering callback methods.
		getLoaderManager().initLoader(LDR_SESSIONS, null, this);
		getLoaderManager().initLoader(LDR_SESSIONS_TOT, null, this);

		
		ListView listView = (ListView) mAct.findViewById(R.id.listview);
		listView.setAdapter(mAdapterSessions);
		
		listView.setOnItemClickListener(this); 
        listView.setOnCreateContextMenuListener(this);
        
        // Get handles and initialize length and noise totals displayed at the top of the list.
        mLabelTotLen = (TextView) mAct.findViewById(R.id.sessions_list_title_total_length);
        mLabelTotNoise =  (TextView) mAct.findViewById(R.id.sessions_list_title_total_noise);
		String lStr;
		lStr = "Total Len: ...";
		mLabelTotLen.setText(lStr);
		lStr = "Total Noise: ...";
		mLabelTotNoise.setText(lStr);
	} // END onResume()

	public void onItemClick(AdapterView<?> l, View v, int position, long rowId) {
		Log.d(LOGTAG, "onItemClick() pos: " + position + ":rowId:" + rowId);

		mAct.updateTab(getString(R.string.frag_tab2_detail), rowId);
	}


	//////////
	// Context Menu creation listener and callback handler - credit to package com.example.android.notepad;
	/**
	 * Create a context menu to Edit, Delete, or Analyze Audio of a practice session from the list.
	 * credit to: package com.example.android.notepad/NotesList.java
	 */
    @Override
    public void onCreateContextMenu(ContextMenu menu, View view, ContextMenuInfo menuInfo) {
        AdapterView.AdapterContextMenuInfo info;
        try {
             info = (AdapterView.AdapterContextMenuInfo) menuInfo;
        } catch (ClassCastException e) {
            Log.e(LOGTAG, "onCreateContextMenu() unable to access menuInfo", e);
            return;
        }

        Cursor cursor = (Cursor) mAdapterSessions.getItem(info.position);
        if (cursor == null) {
            return; // For some reason the requested item isn't available, do nothing
        }

        // Inflate menu from XML resource
        MenuInflater inflater = mAct.getMenuInflater();
        inflater.inflate(R.menu.sessionslist_context_menu, menu);
        
        // Set the context menu header to the formatted session date stamp
        long ldateSec = cursor.getLong(1); // FIXME: ugly hard-coded constant, pointing to the column in the cursor we want... this should be abstracted...
        String ldateStr = SessionsDB.longDateString(ldateSec);
        menu.setHeaderTitle(ldateStr);         
    } // END onCreateContextMenu()
       
	/**
	 * Respond to context menu input from the user.
	 * credit to: package com.example.android.notepad/NotesList.java
	 */
    @Override
    public boolean onContextItemSelected(MenuItem item) {
        AdapterView.AdapterContextMenuInfo info;
        try {
             info = (AdapterView.AdapterContextMenuInfo) item.getMenuInfo();
        } catch (ClassCastException e) {
            Log.e(LOGTAG, "onContextItemSelected() unable to access menuInfo", e);
            return false;
        }

        Log.d(LOGTAG, "onContextItemSelected() 00 item: " + item);  
        Log.d(LOGTAG, "onContextItemSelected() 02   info: " + info + " i.id: " + info.id);        
        Uri lsessionUri = ContentUris.withAppendedId(SessionsDB.CONTENT_PRACTICES_URI, info.id);
        
        Log.d(LOGTAG, "onContextItemSelected() 04   uri: " + lsessionUri);
		
        switch (item.getItemId()) {
        case R.id.slcontext_open:
        	Log.d(LOGTAG, "onContextItemSelected() 13 OPEN ");
    		mAct.updateTab(getString(R.string.frag_tab2_detail), info.id);
            return true;
        case R.id.slcontext_delete:
        	Log.d(LOGTAG, "onContextItemSelected() 14   DELETE");
        	mAct.getContentResolver().delete(lsessionUri, null, null);
        	Log.d(LOGTAG, "onContextItemSelected() 15   delete finished");
            return true;
        case R.id.slcontext_aa:
        	// re analyze audio for session
    		Intent intent = new Intent(mAct, MPQAudioService.class);
    		intent.setAction(MPQAudioService.ACTION_AA);
    		// Get the corresponding filename, and pass it via "extra-bundle".  
    		Cursor lcur = mAct.getContentResolver().query(lsessionUri, new String[] {SessionsDB.PRACTICE_AUDIO_FILE}, null, null, null);
    		Log.d(LOGTAG, "onContextItemSelected() 17   cur: " + lcur.toString() );
    		String lfileName = lcur.getString(0);
    		Log.d(LOGTAG, "onContextItemSelected() 18   NEW filename: " + lfileName);
    		lcur.close();

    		Bundle lextraArgs = new Bundle();
    		lextraArgs.putString(MPQAudioService.ARG_FILENAME, lfileName);
    		intent.putExtras(lextraArgs);
    		mAct.startService(intent);
    		return true;
        default:
        	Log.d(LOGTAG, "onContextItemSelected() 18   ");
        	return super.onContextItemSelected(item);
        }
    } // END onContextItemSelected()

	
	
	// ///////////////////////////////////////////////////
	// LoaderManager.LoaderCallbacks<Cursor> implementations -- FIXME: another very warty area - with a lot of code repeated from the DetailFragment...
	/**
	 * LoaderManager callback to create new loaders.
	 */
	public Loader<Cursor> onCreateLoader(int id, Bundle args) {
		Log.d(LOGTAG, "onCreateLoader() 00 id: " + id);

		String[] lproj;
		String lselc = null;
		String[] lsela = { "" };
		Uri luri;
		
		// Setup projection		
		if (LDR_SESSIONS == id) {
			Log.d(LOGTAG, "onCreateLoader() 14 id: " + id);			
			lproj = new String[] { SessionsDB._ID, 
					SessionsDB.PRACTICE_DATE_STAMP, 
					SessionsDB.PRACTICE_LENGTH_TOTAL,
					SessionsDB.PRACTICE_LENGTH_NOISE};
			luri = SessionsDB.CONTENT_PRACTICES_URI;
		} else { //  if (LDR_SESSIONS_TOT == id) {
			Log.d(LOGTAG, "onCreateLoader() 14 id: " + id);			
			lproj = new String[] { SessionsDB.PRACTICE_LENGTH_TOTAL,
					SessionsDB.PRACTICE_LENGTH_NOISE};
			luri = SessionsDB.CONTENT_PRACTICES_TOTALS_URI;			
		}

		Log.d(LOGTAG, "onCreateLoader() 20 id: " + id);		
		// Get cursor
		CursorLoader lcursLoader = new CursorLoader(mAct,
				luri, lproj, lselc, lsela, null);
		Log.d(LOGTAG, "onCreateLoader() 99");
		return lcursLoader;
	} // END onCreateLoader()

	/**
	 * LoaderManager callback to process when loading cursor data is completed.
	 */
	public void onLoadFinished(Loader<Cursor> loader, Cursor data) {
		int id = loader.getId();
		Log.d(LOGTAG, "onLoadFinished() 00 id: " + id);

		if (null == data) {
			Log.e(LOGTAG, "onLoadFinished() 02 - data is null - return now");
			return;
		}
		if (LDR_SESSIONS == id) {
			mAdapterSessions.swapCursor(data); 
		} else if (LDR_SESSIONS_TOT == id) {
			if (1 == data.getCount()) {
				Log.w(LOGTAG, "onLoadFinished() 11 cursor: " + data.toString() );

				long ltotal_length = data.getLong(0);
				long ltotal_noise = data.getLong(1);
				Log.w(LOGTAG, "onLoadFinished() 13  tlen: " + ltotal_length + " tnoise: " + ltotal_noise);

				String lStr;
				lStr = "Total Len: " + SessionsDB.longSessionMSecString(ltotal_length);
				mLabelTotLen.setText(lStr);

				lStr = "Total Noise: " + SessionsDB.longSessionMSecString(ltotal_noise);
				mLabelTotNoise.setText(lStr);
			}
	//		lcursor.close();        

		} else {
			Log.e(LOGTAG, "onLoadFinished() 66 UNHANDLED ID: " + id);
		}

		Log.w(LOGTAG, "onLoadFinished() 99");
		// The list should now be shown. -- FIXME: is this "now be shown"
		// necessary???
		// if (isResumed()) {
		// setListShown(true);
		// } else {
		// setListShownNoAnimation(true);
		// }
	} // END onLoadFInished()

	/**
	 * LoaderManager callback when releasing data.
	 */
	public void onLoaderReset(Loader<Cursor> loader) {
		Log.d(LOGTAG, "onLoaderReset() 00");
		mAdapterSessions.swapCursor(null);
	}


} // END class SessionsListFragment
